/*
 *  lib-io-tree-utils.h
 *  lib-io-tree
 *
 *  Created by defbond on 8/10/10.
 *  Copyright 2010 LIP/ENS-Lyon. All rights reserved.
 *
 */
#ifndef LIB_IO_TREE_UTILS_H
#define LIB_IO_TREE_UTILS_H

#ifdef __cplusplus

#include <ostream>
#include <iostream>
#include <stdio.h>
#include <list>

#include <vector>
#include <assert.h>
#include <map>

#ifndef DEBUG_MEMUSAGE
#define DEBUG_MEMUSAGE 0 
#endif
#ifndef VERBOSE
#define VERBOSE 0
#endif
#ifndef STRONG_ASSERT
#define STRONG_ASSERT 0
#endif


using namespace std;

#ifndef MAX_COMBI_SIZE
#define MAX_COMBI_SIZE 5
#endif

typedef enum {FURTHEST_NODE=1, BEST_K_COMBI, BEST_FIT_ABS, FIRST_FIT_ABS, BEST_FIT, FIRST_FIT,BEST_INC_COMBI, BEST_COMBI} io_method_t;


double u_wseconds(void);

class Cnode{
  protected:
    bool cost_computed;
    double cost;
    double edge_weight;
    double node_weight;
    vector<Cnode*> * children;
    Cnode * parent;
    unsigned int parent_id;
    unsigned int id;
    int label;

  public :
    Cnode(){
      id = 0;
      Mpeak = 0;
      parent_id = 0;
      Mavail = 0;
      parent = 0;
      cost_computed = false;
      children = new vector<Cnode*>(); 
    }
    Cnode(double nw, double ew){
      id = 0;
      Mpeak = 0;
      parent_id = 0;
      Mavail = 0;
      parent=0;
      cost_computed = false;
      children = new vector<Cnode*>(); 

      edge_weight =ew;
      node_weight = nw;
    }
    Cnode(unsigned int pparent_id,double nw, double ew){
      id = 0;
      Mpeak = 0;
      Mavail = 0;
      parent=0;
      cost_computed = false;
      children = new vector<Cnode*>(); 

      edge_weight =ew;
      node_weight = nw;
      parent_id = pparent_id;
    }

    ~Cnode(){

      for(vector<Cnode*>::iterator iter = children->begin();iter!=children->end();iter++){
        delete *iter;
      }	
      delete children;
    }


    void SetParent(Cnode * pparent){
      this->parent =pparent;
    }


    void AddChild(Cnode * pchild){
      this->children->push_back(pchild);
      cost_computed= false;
    }

    vector<Cnode*> * GetChildren(){
      return children;
    }

    Cnode * GetChild(unsigned int node_id){
      return children->at(node_id);
    }

    Cnode * GetParent(){
      return parent;
    }


    bool IsLeaf() const{
      return children->size()==0;
    }

    bool IsRoot() const{
      return parent_id==0;
    }



    double GetCost(){
      if(!cost_computed){
        //			if(id==308){
        //				cerr<<"EW = "<<edge_weight<<" NW = "<<node_weight<<" children = {";
        //			}
        cost = edge_weight + node_weight;
        for(vector<Cnode*>::iterator iter=children->begin();iter!=children->end();iter++){
          //			if(id==308){
          //				cerr<<(*iter)->GetId()<<"("<<(*iter)->GetEW()<<") ";
          //			}
          cost += (*iter)->GetEW();
        }
        //			if(id==308){
        //			cerr<<"} cost = "<<cost -(edge_weight + node_weight)<<endl;
        //			}
        cost_computed = true;
      }
      return cost;
    }

    void SetParentId(unsigned int pparent_id){
      parent_id = pparent_id;
    }

    void SetId(unsigned int pid){
      id = pid;
    }

    void SetLabel(int pid){
      label = pid;
    }



    void SetEW(double ew ){
      edge_weight = ew;
    }

    void SetNW(double nw ){
      node_weight = nw;
    }

    unsigned int GetParentId() const{
      return parent_id;
    }

    double GetEW() const{
      return edge_weight;
    }

    double GetNW() const{
      return node_weight;
    }

    unsigned int GetId() const{
      return id;
    }

    int GetLabel() const{
      return label;
    }

    void Print(ostream & out) const{
      out<<max((unsigned int)0,GetParentId())<<" "<<GetNW()<<" "<<GetEW()<<endl;
      for(vector<Cnode*>::iterator iter=children->begin();iter!=children->end();iter++){
        (*iter)->Print(out);
      }

    }


    unsigned int Ci;
    double Mpeak;
    double Mavail;
};




class Ctree{
  protected:
    vector<Cnode*> * nodes;
    unsigned int root_index;
    unsigned int root_count;
    unsigned int offset_id;


  public:

    Ctree(){
      root_index=0;
      root_count=0;
      offset_id = 0;
      nodes = new vector<Cnode*>();
    }



    Ctree(int N, int *prnts, double *nwghts, double *ewghts){
      root_index=1;
      root_count=0;
      offset_id = 0;
      nodes = new vector<Cnode*>();

      this->AllocateNodes(N);

      for(int i = 1; i < N+1 ; i++){
        Cnode * cur_node = this->GetNode(i);
        cur_node->GetChildren()->clear();
        cur_node->SetEW(ewghts[i]);
        cur_node->SetNW(nwghts[i]);
        cur_node->SetId(i);
        cur_node->SetLabel(i);
      }

      for(int i = 1; i <N+1 ; i++){
        Cnode * cur_node = this->GetNode(i);

        if(prnts[i] > 0){
          cur_node->SetParentId(prnts[i]);
          cur_node->SetParent(this->GetNode(prnts[i]));
          this->GetNode(prnts[i])->AddChild(cur_node);
        }
        else{
          cur_node->SetParentId(prnts[i]);
          this->SetRootId(i);
        }
      }
    }


    ~Ctree(){
      if(root_index!=0 && nodes->size()>0){
        delete GetRoot();
      }

      delete nodes;
    }

    //TODO recoder ca en recursif
    void Print(ostream & out) const{
      //		if(root_index!=0 && nodes->size()>0){
      //			out<<nodes->size()<<endl;
      //			
      //			GetRoot()->Print(out);
      //		}
      //		

      out<<nodes->size()<<endl;

      for(vector<Cnode*>::iterator iter = nodes->begin();iter!=nodes->end();iter++){
        out<<max((unsigned int)0,(*iter)->GetParentId()/*+1-offset_id*/)<<" "<<(*iter)->GetNW()<<" "<<(*iter)->GetEW()<<endl;
      }

    }


    void AllocateNodes(int new_node_count){
      if(root_count>0 && nodes->size()>0){
        delete GetRoot();
      }

      nodes->resize(new_node_count);

      unsigned int i = 0;
      for(vector<Cnode*>::iterator iter = nodes->begin();iter!=nodes->end();iter++){
        *iter = new Cnode();
        (*iter)->SetId(i++);

      }	


      offset_id = nodes->front()->GetId();
    }

    void AddNode(Cnode * newNode){
      nodes->push_back(newNode);
    }

    void AddRoot(Cnode * newNode){
      root_count++;
      assert(root_count == 1);
      nodes->push_back(newNode);
      root_index = nodes->size()-1;
    }

    Cnode * GetRoot() const{
      return nodes->at(root_index-1);
    }

    unsigned int GetRootId() const{
      return root_index;
    }

    void SetRootId(unsigned int root_id){
      root_index = root_id;
    }


    Cnode * GetNode(unsigned int node_id) const{
      return nodes->at(node_id-1);
    }

    Cnode * GetNodeByPos(unsigned int node_idx) const{
      return nodes->at(node_idx);
    }

    const vector<Cnode*> * GetNodes() const{
      return nodes;
    }
};


struct s_node_t {
  int parent;
  vector<int> children;
  double edge_weight;
  double node_weight;
  double Mpeak;
  double Mavail;
};

struct s_io_t {
  unsigned int node;
  double unloaded_data;
};



typedef list<int> schedule_t;
typedef list<Cnode*> cut_t;

typedef pair<unsigned int, double> io_t;
typedef map<unsigned int, double> io_map;





struct OrdoLiu_t ;
struct val_seg_t {
  schedule_t::iterator begin;
  unsigned int begin_index;
  schedule_t::iterator end;
  unsigned int end_index;
  OrdoLiu_t * orig_ordo;
  double value;
};

struct OrdoLiu_t {
  double max_pebble_cost;
  double fi;
  list<val_seg_t> val_seg;
  schedule_t schedule;
}; 


void ConvertToLiu(const Ctree * tree_us, Ctree * tree_liu) ;
void ConvertToLiu(const int * oldprnts,const double * oldnwghts,const double * oldewghts, int N,const int* chstart,const int * children, int ** pprnts, double ** pnwghts, double ** pewghts);
void parse_tree(const char *filename,Ctree * tree);
void parse_tree(const char *filename,int * N ,int **prnts,double **nwghts,double **ewghts);



extern "C" {
#endif
  void po_construct(const int N, const int * prnts, int **chstart,int **chend,int **children, int * root);
  void poaux(const int * chstart, const int * children, int N, int r, int * por, int * label);
#ifdef __cplusplus
} /* closing brace for extern "C" */



bool check_schedule(int * prnts,int * sched,int N);

double MaxOutDegree(Ctree * tree,int quiet);
double MaxOutDegree(int N, double * nwghts, double * ewghts, int * chstart,int * children);

void NextValley(Cnode * node, double available_memory,  double & cut_value, list<Cnode*> & min_sub_cut, list<unsigned int> & sub_schedule, double & Inc, int quiet, int depth,int & count);
double IOCounter(Ctree & tree, schedule_t & sub_schedule, double available_memory,bool divisible,int quiet);
double IOCounter(int N, double * nwghts, double * ewghts, int * chstart,int * children, int * schedule, double available_memory,bool divisible,int quiet, io_method_t method=FURTHEST_NODE);



#endif
#endif
