#include <cstdlib>

#include <iostream>
#include "lib-io-tree.h"
#include <sys/time.h>
#include <getopt.h>

int main (int argc, char **argv) {
  int quiet = 0;
  int io = 0;
  bool divisible = false;
  int printsched = 0;
  //int count = 0;
  //long int usec;
  double usec;
  bool run_minmem=false;
  bool run_minmema=false;
  bool run_po=false;
  bool run_poi=false;
  bool run_liu=false;
  bool run_liui=false;
  int c;
  while (1) {
    int option_index = 0;
    static struct option long_options[] = {
      {"io", 1, 0, 0},
      {"quiet", 0, 0, 0},
      {"sched", 0, 0, 0},
      {"minmem", 0, 0, 0},
      {"minmemarray", 0, 0, 0},
      {"po", 0, 0, 0},
      {"liu", 0, 0, 0},
      {"poiter", 0, 0, 0},
      {"liuiter", 0, 0, 0},
      {0, 0, 0, 0}
    };

    c = getopt_long (argc, argv, ":i:qsmnplbk",
        long_options, &option_index);
    if (c == -1)
      break;

    switch (c) {
      case 'i':
        {
          io = 1;
          int opt = atoi(optarg);
          divisible = (opt==1);
        }
        break;
      case 'q':
        {
          quiet = 1;
        }
        break;
      case 's':
        {
          printsched=1;
        }
        break;

      case 'm':
        run_minmem=true;
        break;
      case 'n':
        run_minmema=true;
        break;
      case 'p':
        run_po=true;
        break;
      case 'b':
        run_poi=true;
        break;


      case 'l':
        run_liu=true;
        break;


      case 'k':
        run_liui=true;
        break;




      case '?':
        printf("Usage call-io-tree -i (or --io) [0|1] -q (or --quiet) -s (or --sched)  TREE\n");
        break;

      default:
        break;
    }
  }


  if (optind >= argc) {
    printf("YOU MUST PROVIDE A TREE\n");
    exit(1);
  }

  if(!run_minmem && !run_minmema && !run_po && !run_liu && !run_poi && !run_liui){
    run_minmem=true;
  }


  if (!quiet){
    fprintf(stderr,"processing tree \"%s\"\n",argv[optind]);
  }







  double Mr;
  int N;
  int * prnts;
  double * ewghts;
  double * nwghts;

  parse_tree(argv[optind],&N,&prnts,&nwghts,&ewghts);

  int * chstart,*chend,*children;
  int root;

  po_construct(N, prnts, &chstart,&chend,&children, &root);
  double Mp = MaxOutDegree(N, nwghts, ewghts, chstart,children);


  //	Ctree * treeobj = new Ctree(N,prnts,nwghts,ewghts);
  //	double Mp = MaxOutDegree(treeobj,1);
  //	delete treeobj;

  if(run_liui){
    int * schedule = new int[N+1];

    Mr = PebbleOrderingIterAlgorithm_timed(N, prnts, nwghts, ewghts, schedule, &usec,quiet);

    cout<<"*******************"<<endl<<check_schedule(prnts,schedule,N)<<endl<<"********************"<<endl;

    cout<<"***** Iterative PebbleOrdering *****"<<endl;
    if(printsched){
      cout<<"Schedule : {";
      for (int i = 0;i<N;i++){ cout<<schedule[i]<<" "; }
      cout<<"}"<<endl;
    }

    cout<<"Max out deg = "<<Mp<<endl;
    //		if(io)
    //			cerr<<"IO Volume "<<IOCounter(*treeobj, sub_sched, Mp, divisible,quiet)<<endl;
    if(io){
      cout<<"IO Volume "<<IOCounter(N, nwghts, ewghts,chstart,children, schedule, Mp, divisible, quiet)<<endl;
      cout<<"IO Volume "<<IOCounter(N, nwghts, ewghts,chstart,children, schedule, Mp + (Mr-Mp)/2.0, divisible, quiet)<<endl;
      cout<<"IO Volume "<<IOCounter(N, nwghts, ewghts,chstart,children, schedule, Mr, divisible, quiet)<<endl;
    }	
    cout<<"Minimum memory : "<<Mr<<endl;
    cout<<"Answer returned in "<<usec<<" us"<<endl<<endl<<endl;

    delete[] schedule;

  }

  if(run_poi){
    int * schedule = new int[N+1];

    Mr = PostOrderIterAlgorithm_timed(N, prnts, nwghts, ewghts, schedule,  &usec,quiet);

    cout<<"*******************"<<endl<<check_schedule(prnts,schedule,N)<<endl<<"********************"<<endl;

    cout<<"***** Iterative PostOrder *****"<<endl;
    if(printsched){
      cout<<"Schedule : {";
      for (int i = 0;i<N;i++){ cout<<schedule[i]<<" "; }
      cout<<"}"<<endl;
    }

    cout<<"Max out deg = "<<Mp<<endl;
    //		if(io)
    //			cerr<<"IO Volume "<<IOCounter(*treeobj, sub_sched, Mp, divisible,quiet)<<endl;
    if(io){
      cout<<"IO Volume "<<IOCounter(N, nwghts, ewghts,chstart,children, schedule, Mp, divisible, quiet)<<endl;
      cout<<"IO Volume "<<IOCounter(N, nwghts, ewghts,chstart,children, schedule, Mp + (Mr-Mp)/2.0, divisible, quiet)<<endl;
      cout<<"IO Volume "<<IOCounter(N, nwghts, ewghts,chstart,children, schedule, Mr, divisible, quiet)<<endl;
    }
    cout<<"Minimum memory : "<<Mr<<endl;
    cout<<"Answer returned in "<<usec<<" us"<<endl<<endl<<endl;

    delete[] schedule;
  }
  if(run_minmema){
    int * schedule = new int[N+1];

    Mr = MinMemArrayRecurAlgorithm_timed( N, prnts, nwghts, ewghts, schedule, &usec,quiet);

    cout<<"*******************"<<endl<<check_schedule(prnts,schedule,N)<<endl<<"********************"<<endl;

    cout<<"***** MinMem Array *****"<<endl;
    if(printsched){
      cout<<"Schedule : {";
      for (int i = 0;i<N;i++){ cout<<schedule[i]<<" "; }
      cout<<"}"<<endl;
    }

    cout<<"Max out deg = "<<Mp<<endl;
    //		if(io)
    //			cerr<<"IO Volume "<<IOCounter(*treeobj, sub_sched, Mp, divisible,quiet)<<endl;
    if(io){
      cout<<"Mavail = "<<Mp<<" FURTHEST_NODE IO Volume "<<IOCounter(N, nwghts, ewghts,chstart,children, schedule, Mp, divisible, quiet,FURTHEST_NODE)<<endl;
      cout<<"Mavail = "<<Mp<<" BEST_K_COMBI IO Volume "<<IOCounter(N, nwghts, ewghts,chstart,children, schedule, Mp, divisible, quiet,BEST_K_COMBI)<<endl;
//      cout<<"Mavail = "<<Mp<<" BEST_COMBI IO Volume "<<IOCounter(N, nwghts, ewghts,chstart,children, schedule, Mp, divisible, quiet,BEST_COMBI)<<endl;
//      cout<<"Mavail = "<<Mp<<" BEST_INC_COMBI IO Volume "<<IOCounter(N, nwghts, ewghts,chstart,children, schedule, Mp, divisible, quiet,BEST_INC_COMBI)<<endl;
      cout<<"Mavail = "<<Mp<<" BEST_FIT IO Volume "<<IOCounter(N, nwghts, ewghts,chstart,children, schedule, Mp, divisible, quiet,BEST_FIT)<<endl;
      cout<<"Mavail = "<<Mp<<" BEST_FIT_ABS IO Volume "<<IOCounter(N, nwghts, ewghts,chstart,children, schedule, Mp, divisible, quiet,BEST_FIT_ABS)<<endl;
      cout<<"Mavail = "<<Mp<<" FIRST_FIT IO Volume "<<IOCounter(N, nwghts, ewghts,chstart,children, schedule, Mp, divisible, quiet,FIRST_FIT)<<endl;
      cout<<"Mavail = "<<Mp<<" FIRST_FIT_ABS IO Volume "<<IOCounter(N, nwghts, ewghts,chstart,children, schedule, Mp, divisible, quiet,FIRST_FIT_ABS)<<endl;
      cout<<"IO Volume "<<IOCounter(N, nwghts, ewghts,chstart,children, schedule, Mp + (Mr-Mp)/2.0, divisible, quiet)<<endl;
      cout<<"IO Volume "<<IOCounter(N, nwghts, ewghts,chstart,children, schedule, Mr, divisible, quiet)<<endl;
    }

    cout<<"Minimum memory : "<<Mr<<endl;
    cout<<"Answer returned in "<<usec<<" us"<<endl<<endl<<endl;

    delete[] schedule;

  }


  if(run_minmem){
    int * schedule = new int[N+1];

#ifdef DEBUG_USING_MINMEM
    Mr = MinMemRecurAlgorithm_timed(N, prnts, nwghts, ewghts, schedule, &usec,quiet,NULL);
#else
    Mr = MinMemRecurAlgorithm_timed(N, prnts, nwghts, ewghts, schedule, &usec,quiet);
#endif
    cout<<"*******************"<<endl<<check_schedule(prnts,schedule,N)<<endl<<"********************"<<endl;

    cout<<"***** MinMem *****"<<endl;
    if(printsched){
      cout<<"Schedule : {";
      for (int i = 0;i<N;i++){ cout<<schedule[i]<<" "; }
      cout<<"}"<<endl;
    }

    cout<<"Max out deg = "<<Mp<<endl;
    //		if(io)
    //			cerr<<"IO Volume "<<IOCounter(*treeobj, sub_sched, Mp, divisible,quiet)<<endl;

    cout<<"Minimum memory : "<<Mr<<endl;
    cout<<"Answer returned in "<<usec<<" us"<<endl<<endl<<endl;

    delete[] schedule;

  }

  if(run_po){
    int * schedule = new int[N+1];
    Mr = PostOrderRecurAlgorithm_timed(N, prnts, nwghts, ewghts, schedule,&usec,quiet);

    cout<<"*******************"<<endl<<check_schedule(prnts,schedule,N)<<endl<<"********************"<<endl;

    cout<<"***** PostOrder *****"<<endl;
    if(printsched){
      cout<<"Schedule : {";
      for (int i = 0;i<N;i++){ cout<<schedule[i]<<" "; }
      cout<<"}"<<endl;
    }

    cout<<"Max out deg = "<<Mp<<endl;
    //		if(io)
    //			cerr<<"IO Volume "<<IOCounter(*treeobj, sub_sched, Mp, divisible,quiet)<<endl;

    cout<<"Minimum memory : "<<Mr<<endl;
    cout<<"Answer returned in "<<usec<<" us"<<endl<<endl<<endl;

    delete[] schedule;
  }

  if(run_liu){
    int * schedule = new int[N+1];
    Mr = PebbleOrderingRecurAlgorithm_timed(N, prnts, nwghts, ewghts, schedule,&usec,quiet);

    cout<<"*******************"<<endl<<check_schedule(prnts,schedule,N)<<endl<<"********************"<<endl;

    cout<<"***** PebbleOrdering *****"<<endl;
    if(printsched){
      cout<<"Schedule : {";
      for (int i = 0;i<N;i++){ cout<<schedule[i]<<" "; }
      cout<<"}"<<endl;
    }

    cout<<"Max out deg = "<<Mp<<endl;
    //		if(io)
    //			cerr<<"IO Volume "<<IOCounter(*treeobj, sub_sched, Mp, divisible,quiet)<<endl;

    cout<<"Minimum memory : "<<Mr<<endl;
    cout<<"Answer returned in "<<usec<<" us"<<endl<<endl<<endl;

    delete[] schedule;
  }

  delete[] chstart;
  delete[] chend;
  delete[] children;

  delete[] prnts;
  delete[] ewghts;
  delete[] nwghts;



  return 0;

}
