/*
 *  lib-io-tree-utils.cpp
 *  lib-io-tree
 *
 *  Created by defbond on 8/10/10.
 *  Copyright 2010 LIP/ENS-Lyon. All rights reserved.
 *
 */
#include <stdio.h>
#include <cstdlib>
#include <cstring>
#include <fstream>
#include <limits>
#include <cmath>

#include "lib-io-tree-utils.h"
#include "lib-io-tree.h"
#include "lib-io-tree-minmem.h"
#include <sys/time.h>


using namespace std;

double u_wseconds(void) {
  struct timeval tp;

  gettimeofday(&tp, NULL);

  return (double) tp.tv_sec + (double) tp.tv_usec / 1000000.0;
};

void parse_tree(const char *filename,Ctree * tree) {
  ifstream OpenFile(filename);
  char begin;
  char cur_char;
  unsigned int line_index = 1;
  bool nodes_cnt_read = false;

  do
  {	
    /*skip commentary lines*/
    begin = OpenFile.peek();
    if(OpenFile.good() ){

      if(begin=='%'){
        do{
          cur_char = OpenFile.get();
        }while(cur_char != '\n' && OpenFile.good());
      }
      else{
        if(!nodes_cnt_read){
          /* get the number of nodes and skip last trailing character*/
          int nb_of_nodes;
          OpenFile >> nb_of_nodes;
          do{cur_char = OpenFile.get();}while(cur_char != '\n' && OpenFile.good());


          //OpenFile.get(cur_char);
          nodes_cnt_read = true;
          /*allocate space for nodes*/
          tree->AllocateNodes(nb_of_nodes);

        }
        else{
          /*parse actual nodes*/
          unsigned int parent;
          double ew, nw;

          OpenFile >> parent >> nw >> ew;
          do{cur_char = OpenFile.get();}while(cur_char != '\n' && OpenFile.good());




          tree->GetNode(line_index-1)->SetParentId(parent-1);
          tree->GetNode(line_index-1)->SetEW(ew);
          tree->GetNode(line_index-1)->SetNW(nw);
          tree->GetNode(line_index-1)->SetId(line_index-1);

          if(parent > 0){
            tree->GetNode(parent-1)->AddChild(tree->GetNode(line_index-1));
          }

          if(parent==0){
            tree->SetRootId(line_index-1);
          }
          line_index++;
        }
      }

    }
  }while(OpenFile.good() );

          OpenFile.close();
}

void parse_tree(const char *filename,int * N ,int **prnts,double **nwghts,double **ewghts){


  ifstream OpenFile(filename);
  char begin;
  char cur_char;
  int line_index = 1;
  bool nodes_cnt_read = false;


  *N = -1;

  do
  {	
    /*skip commentary lines*/
    begin = OpenFile.peek();
    if(OpenFile.good() ){

      if(begin=='%'){
        do{
          cur_char = OpenFile.get();
        }while(cur_char != '\n' && OpenFile.good());
      }
      else{
        if(!nodes_cnt_read){
          /* get the number of nodes and skip last trailing character*/
          int nodes_count;
          OpenFile >> nodes_count;
          *N = nodes_count;
          do{cur_char = OpenFile.get();}while(cur_char != '\n' && OpenFile.good());


          //OpenFile.get(cur_char);
          nodes_cnt_read = true;
          /*allocate space for nodes*/
          *prnts = new int[nodes_count+1];
          *nwghts = new double[nodes_count+1];
          *ewghts = new double[nodes_count+1];
        }
        else{
          /*parse actual nodes*/
          int parent;
          double ew, nw;

          OpenFile >> parent >> nw >> ew;
          do{cur_char = OpenFile.get();}while(cur_char != '\n' && OpenFile.good());
          (*prnts)[line_index] = parent;
          (*nwghts)[line_index] = nw; 
          (*ewghts)[line_index] = ew; 


          line_index++;
        }
      }

    }
  }while(OpenFile.good() );

          OpenFile.close();



}


//void ConvertToLiu(const Ctree * tree_us, Ctree * tree_liu) {
//	
//
////	tree_liu->GetNodes()->assign(tree_us->GetNodes()->begin(),tree_us->GetNodes()->end())
//	tree_liu->AllocateNodes(2*tree_us->GetNodes()->size());	
//	tree_liu->SetRootId(tree_us->GetRootId());
//	
//	unsigned int N =tree_us->GetNodes()->size();
//
//	//copy the tree
//	for(unsigned int node_index=1;node_index< N+1;node_index++){
//		Cnode * us_node = tree_us->GetNode(node_index);
//		Cnode * liu_edgenode = tree_liu->GetNode(node_index);
//		
//		liu_edgenode->SetId(us_node->GetId());
//		liu_edgenode->SetParentId(us_node->GetParentId());
//		liu_edgenode->SetEW(us_node->GetEW());
//		liu_edgenode->SetNW(us_node->GetNW());
//		
//		liu_edgenode->GetChildren()->clear();
//		for(vector<Cnode*>::iterator iter =us_node->GetChildren()->begin();iter!=us_node->GetChildren()->end();++iter){
//			liu_edgenode->AddChild(tree_liu->GetNode((*iter)->GetId()));
//		}
//		
//	}
//	
//	unsigned int last_node = N+1;	
//	for(unsigned int node_index=1;node_index< N+1;node_index++){
//		double ew, nw ;
//		
//		Cnode * liu_edgenode = tree_liu->GetNode(node_index);
//		Cnode * liu_peaknode = tree_liu->GetNode(last_node);
//		
//
//		nw = liu_edgenode->GetNW();
//		ew = liu_edgenode->GetEW();
//
//		liu_peaknode->SetId(last_node);
//		liu_peaknode->SetParentId(liu_edgenode->GetId());
//		liu_peaknode->SetEW(0);
//		
//		double lnw = ew+nw;
//		/*copy and compute weight of children*/
//		for(vector<Cnode*>::iterator iter =liu_edgenode->GetChildren()->begin();iter!=liu_edgenode->GetChildren()->end();++iter){
//			lnw+= (*iter)->GetEW();
//			liu_peaknode->AddChild((*iter));
//			(*iter)->SetParentId(last_node);
//		}
//		liu_peaknode->SetNW(lnw);
//	
//		liu_edgenode->SetEW(0);
//		liu_edgenode->SetNW(ew);
//		liu_edgenode->GetChildren()->clear();
//		liu_edgenode->AddChild(liu_peaknode);
//		
//		last_node++;
//	}
//	
//	
//	
//}

void ConvertToLiu(const Ctree * tree_us, Ctree * tree_liu) {

  tree_liu->AllocateNodes(2*tree_us->GetNodes()->size());


  tree_liu->SetRootId(tree_us->GetRootId());



  unsigned int last_node = tree_us->GetNodes()->size()+1;

  for(unsigned int node_index=1;node_index< tree_us->GetNodes()->size()+1;node_index++){
    double ew, nw ;

    Cnode * us_node = tree_us->GetNode(node_index);
    Cnode * liu_edgenode = tree_liu->GetNode(node_index);
    Cnode * liu_peaknode = tree_liu->GetNode(last_node);

    liu_edgenode->SetId(node_index);
    liu_peaknode->SetId(last_node);

    nw = us_node->GetNW();
    ew = us_node->GetEW();

    liu_peaknode->SetParentId(liu_edgenode->GetId());
    liu_peaknode->SetEW(0);
    double lnw = ew+nw;
    /*copy and compute weight of children*/
    for(unsigned int j =0;j<us_node->GetChildren()->size();j++){
      lnw+= us_node->GetChild(j)->GetEW();
      liu_peaknode->AddChild(tree_liu->GetNode(us_node->GetChild(j)->GetId()));
      tree_liu->GetNode(us_node->GetChild(j)->GetId())->SetParentId(liu_peaknode->GetId());
    }
    liu_peaknode->SetNW(lnw);

    liu_edgenode->SetParentId(us_node->GetParentId());
    liu_edgenode->AddChild(liu_peaknode);
    liu_edgenode->SetEW(0);
    liu_edgenode->SetNW(ew);

    last_node++;
  }
}


void ConvertToLiu(const int * oldprnts,const double * oldnwghts,const double * oldewghts, int N,const int* chstart,const int * children, int ** pprnts, double ** pnwghts, double ** pewghts) {


  *pprnts = new int[2*N+1];
  memcpy(*pprnts,oldprnts,(N+1)*sizeof(int));
  *pnwghts = new double[2*N+1];
  *pewghts = new double[2*N+1];

  //double max_weight = 0;
  unsigned int last_node = N+1;
  for(int node_index=1;node_index< N + 1;node_index++){
    double ew, nw ;

    int liu_enode = node_index;
    int liu_pnode = last_node;


    nw = oldnwghts[node_index];
    ew = oldewghts[node_index];

    (*pprnts)[liu_pnode] = liu_enode;

    double lnw = ew+nw;
    /*copy and compute weight of children*/
    for(int j = chstart[node_index]; j<chstart[node_index+1];j++){
      int ch = children[j];
      lnw+=oldewghts[ch];
      (*pprnts)[ch] = liu_pnode;
    }

    (*pewghts)[liu_pnode] = 0;
    (*pnwghts)[liu_pnode] = lnw;
    //max_weight=max(max_weight,lnw);
    (*pewghts)[liu_enode] = 0;
    (*pnwghts)[liu_enode] = ew;

    //if(liu_enode==308){
    //cerr<<liu_pnode<<"("<< (*pnwghts)[liu_pnode]<<") >> "<<liu_enode<<"("<<(*pnwghts)[liu_enode]<<") children = {";
    //		for(int j = chstart[node_index]; j<chstart[node_index+1];j++){
    //			int ch = children[j];
    //			cerr<<ch<<"("<<oldewghts[ch]<<") ";
    //		}
    //cerr<<"}"<<endl;
    //}
    last_node++;
  }
  //cerr<<"max "<<max_weight<<endl;
}

void poaux(const int * chstart, const int * children, int N, int r, int * por, int *label){
  int * stack = new int[N+2];
  memset(stack,0,(N+2)*sizeof(*stack));

  int push = 1;
  stack[push]=r;
  push++;
  int * chtmp = (int *)malloc((N+2)*sizeof(int));
  memcpy(chtmp,chstart,(N+2)*sizeof(*chstart));


  while(push>1){
    int top = stack[push-1];
    if(chstart[top+1] - chtmp[top]>0){
      int ch = children[chtmp[top]];
      chtmp[top]++;
      stack[push] = ch;
      push++;
    }
    else
    {
      por[*label] = top;
      (*label)++;
      push--;
    }
  }

  delete[] chtmp;
  delete[] stack;
}



void po_construct(const int N, const int * prnts, int **chstart,int **chend,int **children, int * root){
  *chend = new int[N+2];
  *chstart = new int[N+2];
  *children = new int[N+1];
  memset ( (void *) *chstart, 0, (N+2)*sizeof(**chstart) );
  memset ( (void *) *children, 0, (N+1)*sizeof(**children) );

  *root = -1;
  for(int ii=1;ii<N+1;ii++){
    if(prnts[ii] > 0){
      (*chstart)[prnts[ii]]++;
    }
    else{
      *root = ii;
    }
  }

  /*compute cumsum*/
  int cum_val =1;
  for(int ii=1;ii<N+2;ii++){
    int val = cum_val;
    cum_val+= (*chstart)[ii];
    (*chstart)[ii] = val;
  }


  memcpy(*chend,*chstart,(N+2)*sizeof(**chstart));

  for(int ii=1;ii<N+1;ii++){
    if(prnts[ii] > 0){
      (*children)[(*chend)[prnts[ii]]] = ii;
      (*chend)[prnts[ii]]++;
    }
  }

}


double IOCounter(Ctree & tree, schedule_t & sub_schedule, double available_memory,bool divisible,int quiet){
  double memory_occupation = 0;
  double io_volume = 0;
  io_map unloaded_nodes;	
  schedule_t loaded_nodes;

  /*iterates through the given permutation (schedule)*/
  for(schedule_t::iterator cur_task_id = sub_schedule.begin();cur_task_id!=sub_schedule.end();cur_task_id++){
    Cnode * cur_node = tree.GetNode(*cur_task_id);


    /*if the node was unloaded*/
    if(unloaded_nodes.find(*cur_task_id)!=unloaded_nodes.end()){
      if(!quiet){cerr<<"Loading "<< unloaded_nodes[*cur_task_id] <<"of "<<*cur_task_id<<" (IO)"<<endl;}

      memory_occupation += unloaded_nodes[*cur_task_id];
      unloaded_nodes.erase(*cur_task_id);
    }



    double data_to_unload = memory_occupation + cur_node->GetCost() - cur_node->GetEW() - available_memory;
    if(!quiet){cerr<<"min data to unload "<<data_to_unload<<endl;}
    if(data_to_unload > 0 )
    {
      /*if we dont have enough room, unload files and update both io and occupation*/
      double unloaded_data = 0;
      /*unload furthest non unloaded node which is NOT in current_node children first*/
      schedule_t::reverse_iterator far_node_id = loaded_nodes.rbegin();
      while((far_node_id!=loaded_nodes.rend()) && (unloaded_data < data_to_unload) ){
        /*try to unload this node*/
        bool is_already_unloaded = false;
        double remaining_loaded_data = tree.GetNode(*far_node_id)->GetEW();
        if(unloaded_nodes.find(*far_node_id)!=unloaded_nodes.end()){
          is_already_unloaded = true;
          remaining_loaded_data = max(remaining_loaded_data - unloaded_nodes[*far_node_id],0.0);
        }

        double local_data_to_unload;
        if(divisible){
          local_data_to_unload = min(remaining_loaded_data, data_to_unload);
        }
        else{
          local_data_to_unload = remaining_loaded_data;
        }

        if(!quiet){cerr<<"unloading (IO) "<<local_data_to_unload<<" of "<<*far_node_id<<endl;}
        unloaded_data += local_data_to_unload;
        if(is_already_unloaded){
          unloaded_nodes[*far_node_id]+=local_data_to_unload;
        }
        else{
          unloaded_nodes[*far_node_id] = local_data_to_unload;
        }

        if(remaining_loaded_data  == local_data_to_unload){
          loaded_nodes.remove(*far_node_id);
        }

        far_node_id++;							
      }

      io_volume+=unloaded_data;
      memory_occupation -= unloaded_data;
    }

    if(!quiet){cerr<<"occupation before processing "<<memory_occupation<<endl;}
    /*if we have enough memory to process the node, update occupation*/
    memory_occupation += cur_node->GetCost() - 2*cur_node->GetEW() - cur_node->GetNW();
    if(!quiet){cerr<<"processing "<<*cur_task_id<<endl;}
    if(!quiet){cerr<<"unloading "<<*cur_task_id<<endl;}
    loaded_nodes.remove(*cur_task_id);

    if(!quiet){cerr<<"loading ";}
    for(vector<Cnode*>::iterator child = cur_node->GetChildren()->begin();child!=cur_node->GetChildren()->end();child++){
      if(!quiet){cerr<<(*child)->GetId()<<" ";}
      loaded_nodes.push_back((*child)->GetId());
    }
    if(!quiet){cerr<<endl;}
    if(!quiet){cerr<<"New occupation after processing "<<memory_occupation<<endl;}
  }

  return io_volume;
  //	cerr<<"IO Volume "<<io_volume<<endl;
}



double unload_furthest_nodes(io_map & unloaded_nodes, schedule_t & loaded_nodes,const double data_to_unload, double * ewghts, bool divisible){
      double unloaded_data = 0.0;
      /*unload furthest non unloaded node which is NOT in current_node children first*/ 
      schedule_t::reverse_iterator far_node_id = loaded_nodes.rbegin();	
      while((far_node_id!=loaded_nodes.rend()) && (unloaded_data < data_to_unload) ){
        /*try to unload this node*/
        bool is_already_unloaded = false;
        double remaining_loaded_data = ewghts[*far_node_id];
        if(unloaded_nodes.find(*far_node_id)!=unloaded_nodes.end()){
          is_already_unloaded = true;
          remaining_loaded_data = max(remaining_loaded_data - unloaded_nodes[*far_node_id],0.0);
        }
        double local_data_to_unload;
        if(divisible){
          local_data_to_unload = min(remaining_loaded_data, data_to_unload);
        }
        else{
          local_data_to_unload = remaining_loaded_data;
        }
#if VERBOSE
        cerr<<"unloading (IO) "<<local_data_to_unload<<" of "<<*far_node_id<<endl;
#endif
        unloaded_data += local_data_to_unload;
        if(is_already_unloaded){		
          unloaded_nodes[*far_node_id]+=local_data_to_unload;
        }					
        else{
          unloaded_nodes[*far_node_id] = local_data_to_unload;
        }					
        if(remaining_loaded_data  == local_data_to_unload){
          loaded_nodes.remove(*far_node_id);					
        }							
        far_node_id++;
      }										

	return unloaded_data;
}

double unload_furthest_first_fit(io_map & unloaded_nodes, schedule_t & loaded_nodes,const double data_to_unload, double * ewghts, bool divisible){
      double unloaded_data = 0.0;
      /*unload furthest non unloaded node which is NOT in current_node children first*/ 
      schedule_t::reverse_iterator far_node_id = loaded_nodes.rbegin();	
      while((far_node_id!=loaded_nodes.rend()) && (unloaded_data < data_to_unload) ){
        /*try to unload this node*/
        bool is_already_unloaded = false;
        double remaining_loaded_data = ewghts[*far_node_id];
        if(unloaded_nodes.find(*far_node_id)!=unloaded_nodes.end()){
          is_already_unloaded = true;
          remaining_loaded_data = max(remaining_loaded_data - unloaded_nodes[*far_node_id],0.0);
        }
        double local_data_to_unload;
        if(divisible){
          local_data_to_unload = min(remaining_loaded_data, data_to_unload);
        }
        else{
          local_data_to_unload = remaining_loaded_data;
        }
#if VERBOSE
        cerr<<"unloading (IO) "<<local_data_to_unload<<" of "<<*far_node_id<<endl;
#endif

        /*if it "fits", that is if the amount of data is lower than what we need to unload*/
        if(local_data_to_unload <= data_to_unload - unloaded_data){
          unloaded_data += local_data_to_unload;
          if(is_already_unloaded){		
            unloaded_nodes[*far_node_id]+=local_data_to_unload;
          }					
          else{
            unloaded_nodes[*far_node_id] = local_data_to_unload;
          }					
          if(remaining_loaded_data  == local_data_to_unload){
            loaded_nodes.remove(*far_node_id);					
          }
        }

        far_node_id++;
      }										


      /*if we did not unloaded enough data, call furthest node*/
      if(unloaded_data<data_to_unload){
          unloaded_data += unload_furthest_nodes(unloaded_nodes, loaded_nodes,data_to_unload - unloaded_data, ewghts, divisible);
      }

      return unloaded_data;
}

double unload_furthest_best_fit(io_map & unloaded_nodes, schedule_t & loaded_nodes,const double data_to_unload, double * ewghts, bool divisible){
  double unloaded_data = 0.0;
  /*unload furthest non unloaded node which is NOT in current_node children first*/ 
  while(unloaded_data < data_to_unload){
    bool is_best_already_unloaded = false;
    double best_remaining_loaded_data = 0.0;
    unsigned int best_candidate = -1;
    double best_candi_score = -1.0;

    schedule_t::reverse_iterator far_node_id = loaded_nodes.rbegin();	
    while((far_node_id!=loaded_nodes.rend())){
      /*try to unload this node*/
      bool is_already_unloaded = false;
      double remaining_loaded_data = ewghts[*far_node_id];
      if(unloaded_nodes.find(*far_node_id)!=unloaded_nodes.end()){
        is_already_unloaded = true;
        remaining_loaded_data = max(remaining_loaded_data - unloaded_nodes[*far_node_id],0.0);
      }
      double local_data_to_unload;
      if(divisible){
        local_data_to_unload = min(remaining_loaded_data, data_to_unload);
      }
      else{
        local_data_to_unload = remaining_loaded_data;
      }
#if VERBOSE
      cerr<<"unloading (IO) "<<local_data_to_unload<<" of "<<*far_node_id<<endl;
#endif

      /*if it "fits", that is if the amount of data is lower than what we need to unload*/
      if(local_data_to_unload <= data_to_unload - unloaded_data){
        if(local_data_to_unload > best_candi_score){
          best_candi_score = local_data_to_unload;
          best_candidate = *far_node_id;
          is_best_already_unloaded = is_already_unloaded;
          best_remaining_loaded_data = remaining_loaded_data;
        }
      }

      far_node_id++;
    }


    if(best_candi_score != -1){
      //cerr<<"best found :"<<best_candidate<<endl;
      unloaded_data += best_candi_score;
      if(is_best_already_unloaded){		
        unloaded_nodes[best_candidate]+=best_candi_score;
      }					
      else{
        unloaded_nodes[best_candidate] = best_candi_score;
      }					
      if(best_remaining_loaded_data  == best_candi_score){
        loaded_nodes.remove(best_candidate);					
      }
    }
    else{
      break;
    }
  }
  /*if we did not unloaded enough data, call furthest node*/
  if(unloaded_data<data_to_unload){
    unloaded_data += unload_furthest_nodes(unloaded_nodes, loaded_nodes,data_to_unload - unloaded_data, ewghts, divisible);
  }


	
	return unloaded_data;
}






double unload_furthest_first_fit_abs(io_map & unloaded_nodes, schedule_t & loaded_nodes,const double data_to_unload, double * ewghts, bool divisible){
      double unloaded_data = 0.0;
      /*unload furthest non unloaded node which is NOT in current_node children first*/ 
      schedule_t::reverse_iterator far_node_id = loaded_nodes.rbegin();	
      while((far_node_id!=loaded_nodes.rend()) && (unloaded_data < data_to_unload) ){
        /*try to unload this node*/
        bool is_already_unloaded = false;
        double remaining_loaded_data = ewghts[*far_node_id];
        if(unloaded_nodes.find(*far_node_id)!=unloaded_nodes.end()){
          is_already_unloaded = true;
          remaining_loaded_data = max(remaining_loaded_data - unloaded_nodes[*far_node_id],0.0);
        }
        double local_data_to_unload;
        if(divisible){
          local_data_to_unload = min(remaining_loaded_data, data_to_unload);
        }
        else{
          local_data_to_unload = remaining_loaded_data;
        }
#if VERBOSE
        cerr<<"unloading (IO) "<<local_data_to_unload<<" of "<<*far_node_id<<endl;
#endif

        /*if it "fits", that is if the amount of data is lower than what we need to unload*/
        if(local_data_to_unload >= data_to_unload - unloaded_data){
          unloaded_data += local_data_to_unload;
          if(is_already_unloaded){		
            unloaded_nodes[*far_node_id]+=local_data_to_unload;
          }					
          else{
            unloaded_nodes[*far_node_id] = local_data_to_unload;
          }					
          if(remaining_loaded_data  == local_data_to_unload){
            loaded_nodes.remove(*far_node_id);					
          }
        }

        far_node_id++;
      }										

      return unloaded_data;
}



double unload_furthest_best_fit_abs(io_map & unloaded_nodes, schedule_t & loaded_nodes,const double data_to_unload, double * ewghts, bool divisible){
  double unloaded_data = 0.0;
  /*unload furthest non unloaded node which is NOT in current_node children first*/ 
  while(unloaded_data < data_to_unload){
    bool is_best_already_unloaded = false;
    double best_remaining_loaded_data = 0.0;
    unsigned int best_candidate = -1;
    double best_candi_score = -1.0;

    schedule_t::reverse_iterator far_node_id = loaded_nodes.rbegin();	
    while((far_node_id!=loaded_nodes.rend())){
      /*try to unload this node*/
      bool is_already_unloaded = false;
      double remaining_loaded_data = ewghts[*far_node_id];
      if(unloaded_nodes.find(*far_node_id)!=unloaded_nodes.end()){
        is_already_unloaded = true;
        remaining_loaded_data = max(remaining_loaded_data - unloaded_nodes[*far_node_id],0.0);
      }
      double local_data_to_unload;
      if(divisible){
        local_data_to_unload = min(remaining_loaded_data, data_to_unload);
      }
      else{
        local_data_to_unload = remaining_loaded_data;
      }
#if VERBOSE
      cerr<<"unloading (IO) "<<local_data_to_unload<<" of "<<*far_node_id<<endl;
#endif

      /*if it "fits", that is if the amount of data is lower than what we need to unload*/
      if(abs(data_to_unload-unloaded_data-local_data_to_unload)<abs(data_to_unload-unloaded_data-best_candi_score)){
          best_candi_score = local_data_to_unload;
          best_candidate = *far_node_id;
          is_best_already_unloaded = is_already_unloaded;
          best_remaining_loaded_data = remaining_loaded_data;
      }

      far_node_id++;
    }


    if(best_candi_score != -1){
      //cerr<<"best found :"<<best_candidate<<endl;
      unloaded_data += best_candi_score;
      if(is_best_already_unloaded){		
        unloaded_nodes[best_candidate]+=best_candi_score;
      }					
      else{
        unloaded_nodes[best_candidate] = best_candi_score;
      }					
      if(best_remaining_loaded_data  == best_candi_score){
        loaded_nodes.remove(best_candidate);					
      }
    }
    else{
      break;
    }
  }

	return unloaded_data;
}











/*
	next_comb(int comb[], int k, int n)
		Generates the next combination of n elements as k after comb

	comb => the previous combination ( use (0, 1, 2, ..., k) for first)
	k => the size of the subsets to generate
	n => the size of the original set

	Returns: 1 if a valid combination was found
		0, otherwise
*/
int next_comb(int comb[], int k, int n) {
  int i = k - 1;
  ++comb[i];

  while ((i > 0) && (comb[i] >= n - k + 1 + i)) {
    --i;
    ++comb[i];
  }

  if (comb[0] > n - k) /* Combination (n-k, n-k+1, ..., n) reached */
    return 0; /* No more combinations can be generated */

  /* comb now looks like (..., x, n, n, n, ..., n).
   *  Turn it into (..., x, x + 1, x + 2, ...) */
  for (i = i + 1; i < k; ++i)
    comb[i] = comb[i - 1] + 1;

  return 1;
}



double unload_best_increasing_combi(io_map & unloaded_nodes, schedule_t & loaded_nodes,const double data_to_unload, double * ewghts, bool divisible, unsigned int init_combi_size, bool quiet){
  assert(!divisible);
  vector<unsigned int> candidates;
  double unloaded_data = 0.0;
  vector<unsigned int> best_combi;
  double best_combi_score = numeric_limits<double>::infinity();
  while(unloaded_data<data_to_unload){
    /*unload furthest non unloaded node which is NOT in current_node children first*/
    candidates.clear();
    schedule_t::reverse_iterator far_node_id = loaded_nodes.rbegin();	
    while((far_node_id!=loaded_nodes.rend()) && (candidates.size() <= init_combi_size) ){
      /*try to unload this node*/
      if(unloaded_nodes.find(*far_node_id)==unloaded_nodes.end()){
        candidates.push_back(*far_node_id);
      }

      far_node_id++;
    }
    /*now we got our max_candidates candidates, compute every permutations*/

    vector<unsigned int> cur_combi;

    int n = candidates.size(); /* The size of the set; for {1, 2, 3, 4} it's 4 */
    for(int k = 1;k<=n;k++){
      /* k is the size of the subsets; for {1, 2}, {1, 3}, ... it's 2 */
      int comb[k]; /* comb[i] is the index of the i-th element in the
                      combination */

      /* Setup comb for the initial combination */
      for (int i = 0; i < k; ++i)
        comb[i] = i;
      /*compute score*/
      cur_combi.clear();
      double cur_unloaded_data = 0.0;
      if(!quiet){cerr<<"cur_combi is [";}
      for (int i = 0; i < k; ++i){
        unsigned int cur_node_id = candidates[comb[i]];
        cur_unloaded_data+= ewghts[cur_node_id];
        cur_combi.push_back(cur_node_id);
        if(!quiet){cerr<<" "<<cur_node_id;}
      }
      if(!quiet){cerr<<"], its score is "<<cur_unloaded_data<<endl;}
      

      if(abs(data_to_unload-unloaded_data-cur_unloaded_data)<abs(data_to_unload-unloaded_data-best_combi_score)){
        if(!quiet){cerr<<"cur_combi is better than prev best "<<cur_unloaded_data<<" vs "<<best_combi_score<<endl;}
        best_combi.assign(cur_combi.begin(),cur_combi.end());
        best_combi_score = cur_unloaded_data;
      }


      /* Generate and print all the other combinations */
      while (next_comb(comb, k, n)){
        /*compute score*/
        cur_combi.clear();
        double cur_unloaded_data = 0.0;
        if(!quiet){cerr<<"cur_combi is [";}
        for (int i = 0; i < k; ++i){
          unsigned int cur_node_id = candidates[comb[i]];
          cur_unloaded_data+= ewghts[cur_node_id];
          cur_combi.push_back(cur_node_id);
          if(!quiet){cerr<<" "<<cur_node_id;}
        }
        if(!quiet){cerr<<"], its score is "<<cur_unloaded_data<<endl;}

        if(abs(data_to_unload-unloaded_data-cur_unloaded_data)<abs(data_to_unload-unloaded_data-best_combi_score)){
          if(!quiet){cerr<<"cur_combi is better than prev best "<<cur_unloaded_data<<" vs "<<best_combi_score<<endl;}
          best_combi.assign(cur_combi.begin(),cur_combi.end());
          best_combi_score = cur_unloaded_data;
        }
      }
    }

    unloaded_data = best_combi_score;
    
    if(unloaded_data < data_to_unload){
      init_combi_size*=2;
    }


  }										
    if(!quiet){cerr<<"best_combi is [";}
    for(unsigned int i=0;i<best_combi.size();i++){
      unsigned int cur_id = best_combi[i];
      unloaded_nodes[cur_id] = ewghts[cur_id];
      loaded_nodes.remove(cur_id);
      if(!quiet){cerr<<" "<<cur_id;}
    }
    if(!quiet){cerr<<"], its score is "<<best_combi_score<<endl;}
    
    if(!quiet){cerr<<"We've unloaded "<<unloaded_data<<" out of "<<data_to_unload<<" so far"<<endl;}
    unloaded_data = best_combi_score;
	return unloaded_data;
}







double unload_best_furthest_nodes(io_map & unloaded_nodes, schedule_t & loaded_nodes,const double data_to_unload, double * ewghts, bool divisible, unsigned int max_candidates,bool quiet){
  assert(!divisible);
  vector<unsigned int> candidates;
  double unloaded_data = 0.0;
  while(unloaded_data<data_to_unload){
    /*unload furthest non unloaded node which is NOT in current_node children first*/
    candidates.clear();
    schedule_t::reverse_iterator far_node_id = loaded_nodes.rbegin();	
    while((far_node_id!=loaded_nodes.rend()) && (candidates.size() <= max_candidates) ){
      /*try to unload this node*/
      if(unloaded_nodes.find(*far_node_id)==unloaded_nodes.end()){
        candidates.push_back(*far_node_id);
      }

      far_node_id++;
    }
    /*now we got our max_candidates candidates, compute every permutations*/



    vector<unsigned int> best_combi;
    vector<unsigned int> cur_combi;
    double best_combi_score = numeric_limits<double>::infinity();

    int n = candidates.size(); /* The size of the set; for {1, 2, 3, 4} it's 4 */
    for(int k = 1;k<=n;k++){
      /* k is the size of the subsets; for {1, 2}, {1, 3}, ... it's 2 */
      int comb[k]; /* comb[i] is the index of the i-th element in the
                      combination */

      /* Setup comb for the initial combination */
      for (int i = 0; i < k; ++i)
        comb[i] = i;
      /*compute score*/
      cur_combi.clear();
      double cur_unloaded_data = 0.0;
      if(!quiet){cerr<<"cur_combi is [";}
      for (int i = 0; i < k; ++i){
        unsigned int cur_node_id = candidates[comb[i]];
        cur_unloaded_data+= ewghts[cur_node_id];
        cur_combi.push_back(cur_node_id);
        if(!quiet){cerr<<" "<<cur_node_id;}
      }
      if(!quiet){cerr<<"], its score is "<<cur_unloaded_data<<endl;}
      

      if((double)abs(data_to_unload-unloaded_data-cur_unloaded_data)<(double)abs(data_to_unload-unloaded_data-best_combi_score)){
        if(!quiet){cerr<<"cur_combi is better than prev best "<<cur_unloaded_data<<" vs "<<best_combi_score<<endl;}
        best_combi.assign(cur_combi.begin(),cur_combi.end());
        best_combi_score = cur_unloaded_data;
      }


      /* Generate and print all the other combinations */
      while (next_comb(comb, k, n)){
        /*compute score*/
        cur_combi.clear();
        double cur_unloaded_data = 0.0;
        if(!quiet){cerr<<"cur_combi is [";}
        for ( int i = 0; i < k; ++i){
          unsigned int cur_node_id = candidates[comb[i]];
          cur_unloaded_data+= ewghts[cur_node_id];
          cur_combi.push_back(cur_node_id);
          if(!quiet){cerr<<" "<<cur_node_id;}
        }
        if(!quiet){cerr<<"], its score is "<<cur_unloaded_data<<endl;}

        if(abs(data_to_unload-unloaded_data-cur_unloaded_data)<abs(data_to_unload-unloaded_data-best_combi_score)){
          if(!quiet){cerr<<"cur_combi is better than prev best "<<cur_unloaded_data<<" vs "<<best_combi_score<<endl;}
          best_combi.assign(cur_combi.begin(),cur_combi.end());
          best_combi_score = cur_unloaded_data;
        }
      }
    }


    if(!quiet){cerr<<"best_combi is [";}
    for(unsigned int i=0;i<best_combi.size();i++){
      unsigned int cur_id = best_combi[i];
      unloaded_nodes[cur_id] = ewghts[cur_id];
      loaded_nodes.remove(cur_id);
      if(!quiet){cerr<<" "<<cur_id;}
    }
    if(!quiet){cerr<<"], its score is "<<best_combi_score<<endl;}

    unloaded_data += best_combi_score;
    if(!quiet){cerr<<"We've unloaded "<<unloaded_data<<" out of "<<data_to_unload<<" so far"<<endl;}

  }										

	return unloaded_data;
}


double IOCounter(int N, double * nwghts, double * ewghts, int * chstart,int * children, int * schedule, double available_memory,bool divisible,int quiet,io_method_t method){
  double memory_occupation = 0;
  double io_volume = 0;
  io_map unloaded_nodes;	
  schedule_t loaded_nodes;

  /*iterates through the given permutation (schedule)*/
  for(int rank = N-1;rank>=1;rank--){
    int cur_task_id = schedule[rank];		

    /*if the node was unloaded*/
    if(unloaded_nodes.find(cur_task_id)!=unloaded_nodes.end()){
      if(!quiet){cerr<<"Loading "<< unloaded_nodes[cur_task_id] <<"of "<<cur_task_id<<" (IO)"<<endl;}

      memory_occupation += unloaded_nodes[cur_task_id];
      unloaded_nodes.erase(cur_task_id);
    }


    double node_cost = ewghts[cur_task_id] + nwghts[cur_task_id];
    for(int j = chstart[cur_task_id];j<chstart[cur_task_id+1];j++){
      node_cost += ewghts[children[j]];	
    }

    double data_to_unload = memory_occupation + node_cost - ewghts[cur_task_id] - available_memory;

    if(!quiet){cerr<<"min data to unload "<<data_to_unload<<endl;}
    if(data_to_unload > 0 )
    {
      //cerr<<"We must commit I/O in order to process node "<<cur_task_id<<" which requires "<< memory_occupation + node_cost - ewghts[cur_task_id]<< " but has "<<available_memory<<"available"<<endl;
      /*if we dont have enough room, unload files and update both io and occupation*/
        double unloaded_data = 0.0;
        if(!divisible){
          switch (method){
            case BEST_K_COMBI:
              unloaded_data = unload_best_furthest_nodes(unloaded_nodes, loaded_nodes,data_to_unload, ewghts, divisible, MAX_COMBI_SIZE,quiet);
              break;
            case BEST_INC_COMBI:
              unloaded_data = unload_best_increasing_combi(unloaded_nodes, loaded_nodes,data_to_unload, ewghts, divisible, MAX_COMBI_SIZE,quiet);
              break;

            case BEST_COMBI:
              unloaded_data = unload_best_furthest_nodes(unloaded_nodes, loaded_nodes,data_to_unload, ewghts, divisible, loaded_nodes.size(),quiet);
              break;
            case FURTHEST_NODE:
              unloaded_data = unload_furthest_nodes(unloaded_nodes, loaded_nodes,data_to_unload, ewghts, divisible);
              break;
            case FIRST_FIT:
              unloaded_data = unload_furthest_first_fit(unloaded_nodes,loaded_nodes,data_to_unload, ewghts, divisible);
              break;
            case BEST_FIT:
              unloaded_data = unload_furthest_best_fit(unloaded_nodes,loaded_nodes,data_to_unload, ewghts, divisible);
              break;
            case FIRST_FIT_ABS:
              unloaded_data = unload_furthest_first_fit_abs(unloaded_nodes,loaded_nodes,data_to_unload, ewghts, divisible);
              break;
            case BEST_FIT_ABS:
              unloaded_data = unload_furthest_best_fit_abs(unloaded_nodes,loaded_nodes,data_to_unload, ewghts, divisible);
              break;
            default:
              unloaded_data = unload_furthest_nodes(unloaded_nodes, loaded_nodes,data_to_unload, ewghts, divisible);
              break;
          }
        }
        else{
          unloaded_data = unload_furthest_nodes(unloaded_nodes, loaded_nodes,data_to_unload, ewghts, divisible);
        }
      io_volume+=unloaded_data;
      memory_occupation -= unloaded_data;
    }

    if(!quiet){cerr<<"occupation before processing "<<memory_occupation<<endl;}
    /*if we have enough memory to process the node, update occupation*/
    memory_occupation += node_cost - 2*ewghts[cur_task_id] - nwghts[cur_task_id];
    //		memory_occupation += node_cost - ewghts[cur_task_id] - nwghts[cur_task_id];
    memory_occupation = max(0.0,memory_occupation);

    if(!quiet){cerr<<"processing "<<cur_task_id<<endl;}
    if(!quiet){cerr<<"unloading "<<cur_task_id<<endl;}
    loaded_nodes.remove(cur_task_id);

    if(!quiet){cerr<<"loading ";}
    for(int j = chstart[cur_task_id];j<chstart[cur_task_id+1];j++){
      int ch = children[j];
      if(!quiet){cerr<<ch<<" ";}
      loaded_nodes.push_back(ch);	
    }

    if(!quiet){cerr<<endl;}
    if(!quiet){cerr<<"New occupation after processing "<<memory_occupation<<endl;}
  }

  return io_volume;
  //	cerr<<"IO Volume "<<io_volume<<endl;
}

double IOCounter(int N, int * prnts, double * nwghts, double * ewghts, int * schedule, double available_memory,int divisible,io_method_t method){
  bool div = (divisible==0);

  int * chstart,*chend,*children;
  int root;

  po_construct(N, prnts, &chstart,&chend,&children, &root);
  double io_volume = IOCounter(N, nwghts, ewghts, chstart,children,schedule, available_memory,div,true,method);


  delete[] chstart;
  delete[] chend;
  delete[] children;

  return io_volume;
  //	cerr<<"IO Volume "<<io_volume<<endl;
}


bool check_schedule(int * prnts,int * sched,int N){
  bool valid = true;

  int * rev_sched = new int[N+1];

  for(int i = 1; i<N+1;i++){
    rev_sched[sched[i-1]] = i;

    //		cerr<<"Task "<<sched[i-1]<<" is scheduled at step "<<i<<endl; 
  }

  for(int i = 1; i<N+1;i++){
    if(prnts[i]>0){
      //			cerr<<"T"<<i<<" at "<<rev_sched[i]<<" || T"<<prnts[i]<<" at "<<rev_sched[prnts[i]]<<endl;
      valid = valid & (rev_sched[prnts[i]]>rev_sched[i]);
      if(rev_sched[prnts[i]]<rev_sched[i]){
        cerr<<"Task "<<prnts[i]<<" is before Task "<<i<<endl;
      }
    }
  }

  delete[] rev_sched;

  return valid;
}

double MaxOutDegree(Ctree * tree,int quiet){
  double max_out =0;
  double max_j = 0;
  for(unsigned int j=1; j<tree->GetNodes()->size(); j++) {
    double cur_out = tree->GetNode(j)->GetCost();	
    if(cur_out>=max_out){
      max_out = cur_out;
      max_j = tree->GetNode(j)->GetId();
    }
  }
  if(!quiet){cerr<<"Max out degree "<<max_out<<" at "<<max_j<<endl;}
  return max_out;
}

double MaxOutDegree(int N, double * nwghts, double * ewghts, int * chstart,int * children){
  double max_out =0;
  //	double leaves =0;
  //	double max_j = 0;
  for(int j=1; j<N+1; j++) {
    double cur_out = nwghts[j]+ewghts[j];
    //         if(chstart[j]==chstart[j+1]){
    //            leaves++;
    //         }

    for(int ch = chstart[j];ch<chstart[j+1];ch++){
      cur_out+=ewghts[children[ch]];
    }

    if(cur_out>=max_out){
      max_out = cur_out;
      //			max_j = j;
    }	
  }
  //cout<<leaves<<" leaves"<<endl;
  //cout<<max_j<<" is the max node"<<endl;
  return max_out;
}

double MaxOutDegree(int N, int * prnts, double * nwghts, double * ewghts){
  int * chstart,*chend,*children;
  int root;

  po_construct(N, prnts, &chstart,&chend,&children, &root);

  double max_out = MaxOutDegree(N, nwghts, ewghts, chstart, children);

  delete[] chstart;
  delete[] chend;
  delete[] children;

  return max_out;
}
